#include "AudioDecoder.h"
#include "AudioDrawer.h"
extern "C"
{
	#include <libavcodec/avcodec.h>
}

AudioDecoder::AudioDecoder()
{

}
AudioDecoder::~AudioDecoder()
{
	if(sampleBufferSet)
		std::free(sampleBuffer);
}

bool AudioDecoder::decodeFile(std::string* filename)
{
	close();

	AVFormatContext* formatCtx = avformat_alloc_context();
	if(avformat_open_input(&formatCtx, filename->c_str(), nullptr, nullptr) < 0)
	{
		wxGetApp().popUpErrorDialog("Couldn't open \"" + *filename + "\".");
		avformat_close_input(&formatCtx);
		return false;
	}

	if(avformat_find_stream_info(formatCtx, nullptr) < 0)
	{
		wxGetApp().popUpErrorDialog("Couldn't find file info for \"" + *filename + "\".");
		avformat_close_input(&formatCtx);
		return false;
	}

	av_dump_format(formatCtx, 0, filename->c_str(), false);

	int streamID = -1;
	for(int i = 0; i < formatCtx->nb_streams; i++)
	{
		if(formatCtx->streams[i]->codec->codec_type == AVMEDIA_TYPE_AUDIO)
		{
			streamID = i;
			break;
		}
	}
	if(streamID == -1)
	{
		wxGetApp().popUpErrorDialog("\"" + *filename + "\" does not contain audio.");
		avformat_close_input(&formatCtx);
		return false;
	}

	AVCodecContext* codecCtx = formatCtx->streams[streamID]->codec;
	AVCodec* codec = avcodec_find_decoder(codecCtx->codec_id);
	if(!codec)
	{
		wxGetApp().popUpErrorDialog("Couldn't find the codec for \"" + *filename + "\".");
		avformat_close_input(&formatCtx);
		return false;
	}
	if(avcodec_open2(codecCtx, codec, nullptr) < 0)
	{
		wxGetApp().popUpErrorDialog("Couldn't open the codec for\"" + *filename + "\".");
		avformat_close_input(&formatCtx);
		return false;
	}

	sampleFormat = codecCtx->sample_fmt;
	if(!(sampleFormat == AV_SAMPLE_FMT_U8 || sampleFormat == AV_SAMPLE_FMT_U8P ||
		sampleFormat == AV_SAMPLE_FMT_S16 || sampleFormat == AV_SAMPLE_FMT_S16P ||
		sampleFormat == AV_SAMPLE_FMT_S32 || sampleFormat == AV_SAMPLE_FMT_S32P ||
		sampleFormat == AV_SAMPLE_FMT_FLT || sampleFormat == AV_SAMPLE_FMT_FLTP ||
		sampleFormat == AV_SAMPLE_FMT_DBL || sampleFormat == AV_SAMPLE_FMT_DBLP))
	{
		wxGetApp().popUpErrorDialog("\"" + *filename + "\" uses an unsupported format.");
		avcodec_close(codecCtx);
		avformat_close_input(&formatCtx);
		return false;
	}

	planar = false;
	if(sampleFormat == AV_SAMPLE_FMT_U8P || sampleFormat == AV_SAMPLE_FMT_S16P ||
	   sampleFormat == AV_SAMPLE_FMT_S32P || sampleFormat == AV_SAMPLE_FMT_FLTP || sampleFormat == AV_SAMPLE_FMT_DBLP)
	{
		planar = true;
	}

	AVFrame* frame = av_frame_alloc();
	if(!frame)
	{
		wxGetApp().popUpErrorDialog("Failed to allocate an audio frame.");
		avcodec_close(codecCtx);
		avformat_close_input(&formatCtx);
		return false;
	}

	AVPacket packet;
	av_init_packet(&packet);

	duration = formatCtx->duration / (double)AV_TIME_BASE;

	uint64_t estimatedBuffSize = std::ceil(duration * codecCtx->sample_rate * av_get_bytes_per_sample(codecCtx->sample_fmt) * codecCtx->channels);

	sampleBuffer = (uint8_t*)std::malloc(estimatedBuffSize);
	sampleBufferSet = true;

	int len;
	int gotFrame = 0;
	uint64_t dataSize;
	uint64_t totalBufferSize = 0;
	uint64_t totalSamples = 0;
	while(av_read_frame(formatCtx, &packet) == 0)
	{
		len = avcodec_decode_audio4(codecCtx, frame, &gotFrame, &packet);
		if(len < 0)
		{
			wxGetApp().popUpErrorDialog("Error while decoding.");
			std::free(sampleBuffer);
			sampleBufferSet = false;
			av_packet_unref(&packet);
			av_frame_free(&frame);
			avcodec_close(codecCtx);
			avformat_close_input(&formatCtx);
			return false;
		}

		if(gotFrame)
		{
			dataSize = av_samples_get_buffer_size(nullptr, codecCtx->channels, frame->nb_samples, codecCtx->sample_fmt, 1);

			while(totalBufferSize + dataSize > estimatedBuffSize)
			{
				estimatedBuffSize *= 1.1;
				sampleBuffer = (uint8_t*)std::realloc(sampleBuffer, estimatedBuffSize);
			}

			std::memcpy(sampleBuffer + totalBufferSize, frame->extended_data[0], dataSize);

			totalBufferSize += dataSize;
			totalSamples += frame->nb_samples;
		}

		av_packet_unref(&packet);
	}

	sampleBuffer = (uint8_t*)std::realloc(sampleBuffer, totalBufferSize);

	numChannels = codecCtx->channels;
	sampleRate = codecCtx->sample_rate;
	numSamples = totalSamples;
	bufferSize = totalBufferSize;

	av_packet_unref(&packet);
	av_frame_free(&frame);
	avcodec_close(codecCtx);
	avformat_close_input(&formatCtx);

	didInit = true;

	return true;
}

void AudioDecoder::close()
{
	if(sampleBufferSet)
	{
		std::free(sampleBuffer);
		sampleBufferSet = false;
	}
	bufferSize = 0;
	numSamples = 0;
	sampleRate = 0;
	numChannels = 0;
	duration = 0;
	sampleFormat = AV_SAMPLE_FMT_NONE;
	planar = false;
	didInit = false;
}

float AudioDecoder::getSample(uint64_t sample, int channel)
{
	float rv = 0.0f;
	if(didInit)
	{
		if(sample > numSamples || channel < 0 || channel >= numChannels)
		{
			return rv;
		}

		switch(sampleFormat)
		{
			case AV_SAMPLE_FMT_U8:
			{
				int8_t tmp = ((int8_t*)sampleBuffer)[numChannels*sample + channel];
				rv = (float)tmp / 255.0f;
				break;
			}
			case AV_SAMPLE_FMT_S16:
			{
				int16_t tmp = ((int16_t*)sampleBuffer)[numChannels*sample + channel];
				rv = (float)tmp / 32767.0f;
				break;
			}
			case AV_SAMPLE_FMT_S32:
			{
				int32_t tmp = ((int32_t*)sampleBuffer)[numChannels*sample + channel];
				rv = (float)tmp / 2147483647.0f;
				break;
			}
			case AV_SAMPLE_FMT_FLT:
			{
				rv = ((float*)sampleBuffer)[numChannels*sample + channel];
				break;
			}
			case AV_SAMPLE_FMT_DBL:
			{
				rv = (float)((double*)sampleBuffer)[numChannels*sample + channel];
				break;
			}
			case AV_SAMPLE_FMT_U8P:
			{
				int8_t tmp = ((int8_t*)sampleBuffer)[sample + numSamples*channel];
				rv = (float)tmp / 255.0f;
				break;
			}
			case AV_SAMPLE_FMT_S16P:
			{
				int16_t tmp = ((int16_t*)sampleBuffer)[sample + numSamples*channel];
				rv = (float)tmp / 32767.0f;
				break;
			}
			case AV_SAMPLE_FMT_S32P:
			{
				int32_t tmp = ((int32_t*)sampleBuffer)[sample + numSamples*channel];
				rv = (float)tmp / 2147483647.0f;
				break;
			}
			case AV_SAMPLE_FMT_FLTP:
			{
				rv = ((float*)sampleBuffer)[sample + numSamples*channel];
				break;
			}
			case AV_SAMPLE_FMT_DBLP:
			{
				rv = (float)((double*)sampleBuffer)[sample + numSamples*channel];
				break;
			}
		}
	}

	return rv;
}

uint8_t* AudioDecoder::getSampleBuffer()
{
	return sampleBuffer;
}
uint64_t AudioDecoder::getBufferSize()
{
	return bufferSize;
}
uint64_t AudioDecoder::getNumSamples()
{
	return numSamples;
}
int AudioDecoder::getSampleRate()
{
	return sampleRate;
}
int AudioDecoder::getNumChannels()
{
	return numChannels;
}
double AudioDecoder::getDuration()
{
	return duration;
}
bool AudioDecoder::getIsPlanar()
{
	return planar;
}
bool AudioDecoder::getDidInit()
{
	return didInit;
}
AVSampleFormat AudioDecoder::getSampleFormat()
{
	return sampleFormat;
}